import re
from typing import Any

import mysql.connector

from .connection_base import ConnectionBase
from .managed_cursor import ManagedCursor


class MySqlConnection(ConnectionBase):
    def __init__(self, connection_string):
        super().__init__(connection_string)
        self.provider_name = "mysql"
        match = re.match(r"mysql:\/\/(\w+):(\w+)@(\w+)\/(\w+)", self.connection_string)
        if match:
            self.user = match.group(1)
            self.password = match.group(2)
            self.hostname = match.group(3)
            self.database = match.group(4)
        else:
            raise Exception("Invalid connection string")

        self.connection = mysql.connector.connect(user=self.user, password=self.password, host=self.hostname,
                                                  database=self.database)
        self.cursor = self.connection.cursor()

    def start(self):
        self.cursor.execute("BEGIN TRANSACTION;", {})

    def commit(self):
        self.cursor.execute("COMMIT;")

    def rollback(self):
        self.cursor.execute("ROLLBACK;")

    def execute(self, query: str, params: None):
        if params is None:
            params = {}
        self.cursor.execute(query, params)

    def execute_lastrowid(self, query: str, params: {}) -> Any:
        if params is None:
            params = {}
        self.cursor.execute(query, params)
        return self.cursor.lastrowid

    def fetch(self, query: str, params=None) -> ManagedCursor:
        if params is None:
            params = {}
        cursor = self.connection.cursor()
        cursor.execute(query, params)
        return ManagedCursor(cursor)

    def close(self):
        self.connection.close()



from pathlib import Path
from typing import List


def get_basename(path: str) -> str:
    p = Path(path)
    p.resolve()
    return p.name


def get_fullname(path: str) -> str:
    p = Path(path)
    p.resolve()
    return str(p.expanduser())


def get_filename(path: str) -> str:
    p = Path(path)
    p.resolve()
    return str(p.expanduser()).replace(str(p.parent), "").replace("/", "")


def split_clean(value: str, sep: str) -> List[str]:
    return list(filter(None, value.split(sep)))


def clean_string(dirty: str) -> str:
    while "\n" in dirty:
        dirty = dirty.replace("\n", " ")

    while "\r" in dirty:
        dirty = dirty.replace("\r", " ")

    while "\t" in dirty:
        dirty = dirty.replace("\t", " ")

    while "  " in dirty:
        dirty = dirty.replace("  ", " ")

    dirty = dirty.strip()
    if dirty[-1] == ",":
        dirty = dirty[:-1]
    return dirty.strip()


def find_in_list(value: str, items: []) -> int:
    for i in range(len(items)):
        if items[i] == value:
            return i
    return -1


def find_first_in(value: str, items: [str]) -> tuple[int, str]:
    for i in range(len(items)):
        if items[i] in value:
            return value.find(items[i]), items[i]
    return -1, ""


def find_first_in_array(value: str, items: [[str]]) -> tuple[int, [str]]:
    for i in range(len(items)):
        if items[i][0] in value:
            return value.find(items[i][0]), items[i][1:]
    return -1, ""


def pascal(value: str) -> str:
    if value is None or value == "":
        return ""

    return value[0].upper() + value[1:].lower()


def safeget(d: dict, k, default):
    return d.get(k) if k in d else default


from .connection_base import ConnectionBase
from .managed_cursor import ManagedCursor


class Session(object):
    def __init__(self, connection: ConnectionBase = None):
        self.connection = connection

    def __enter__(self):
        self.start()
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def close(self):
        self.commit()
        self.connection.close()

    def start(self):
        self.connection.start()

    def commit(self):
        self.connection.commit()
        self.start()

    def rollback(self):
        self.connection.rollback()
        self.start()

    def execute(self, query: str, params=None) -> None:
        self.connection.execute(query, params)

    def execute_lastrowid(self, query: str, params=None):
        return self.connection.execute_lastrowid(query, params)

    def fetch_scalar(self, query: str, params=None):
        if params is None:
            params = {}
        row = self.fetch_one(query, params)
        if row is not None:
            value = row[0]
        else:
            value = None
        return value

    def fetch_one(self, query: str, params=None):
        if params is None:
            params = {}
        cursor = self.connection.new_cursor()
        cursor.execute(query, params)
        return cursor.fetchone()

    def fetch(self, query: str, params=None) -> ManagedCursor:
        return self.connection.fetch(query, params)


class PersistentSession(Session):
    __global_connection__: ConnectionBase = None

    def __init__(self, connection: ConnectionBase = None):
        #  super().__init__() # deliberately not calling this
        if PersistentSession.__global_connection__ is None:
            PersistentSession.__global_connection__ = connection

        self.connection = PersistentSession.__global_connection__
        # self.in_transaction = False
        # self.cursor = self.connection.cursor

    def __exit__(self, type, value, traceback):
        self.connection.commit()

    def close(self):
        pass



import sqlite3
from typing import Any

from .connection_base import ConnectionBase
from .managed_cursor import ManagedCursor
from .utils import get_fullname, get_filename


class SqliteConnection(ConnectionBase):
    def __init__(self, connection_string):
        super().__init__(connection_string)
        self.provider_name = "sqlite"
        connection_string = self.connection_string.replace("sqlite://", "")
        connection_string = get_fullname(connection_string)
        self.connection = sqlite3.connect(connection_string, check_same_thread=False)
        self.connection.isolation_level = None
        self.database = get_filename(connection_string)
        self.cursor = self.connection.cursor()

    def start(self):
        self.cursor.execute("BEGIN TRANSACTION;", {})

    def commit(self):
        self.cursor.execute("COMMIT;")

    def rollback(self):
        self.cursor.execute("ROLLBACK;")

    def execute(self, query: str, params: None):
        if params is None:
            params = {}
        self.cursor.execute(query, params)

    def execute_lastrowid(self, query: str, params: {})->Any:
        if params is None:
            params = {}
        self.cursor.execute(query, params)
        return self.cursor.lastrowid

    def fetch(self, query: str, params=None) -> ManagedCursor:
        if params is None:
            params = {}
        cursor = self.connection.cursor()
        cursor.execute(query, params)
        return ManagedCursor(cursor)

    def close(self):
        self.connection.close()



import re

from .connection_base import ConnectionBase
from .exceptions import DataException
from .mysql_connection import MySqlConnection
from .pgsql_connection import PgSqlConnection
from .session import Session, PersistentSession
from .sqlite_connection import SqliteConnection


class SessionFactory(object):
    @staticmethod
    def get_connection(connection_string: str) -> ConnectionBase:
        match = re.search(r"(\w+):\/\/(.+)", connection_string)
        if match:
            db_type = match.group(1)

            if db_type == "sqlite":
                return SqliteConnection(connection_string)
            elif db_type == "mysql":
                return MySqlConnection(connection_string)
            elif db_type == "pgsql":
                return PgSqlConnection(connection_string)
            else:
                raise DataException("invalid connection string")

    @staticmethod
    def connect(connection_string: str) -> Session:
        if "memory" in connection_string:
            if PersistentSession.__global_connection__ is None:
                session = PersistentSession(SessionFactory.get_connection(connection_string))
            else:
                session = PersistentSession()

        else:
            session = Session(SessionFactory.get_connection(connection_string))

        return session



class ManagedCursor(object):

    def __init__(self, cursor):
        self.cursor = cursor

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.cursor.close()

    def close(self):
        self.cursor.close()

    def execute(self, sql: str, params: None):
        if params is None:
            params = {}
        return self.cursor.execute(sql, params)

    def executemany(self, sql: str, seq_of_parameters):
        return self.cursor.executemany(sql, seq_of_parameters)

    def fetchall(self):
        return self.cursor.fetchall()

    def fetchone(self):
        return self.cursor.fetchone()

    def __iter__(self):
        return self.cursor.__iter__()

    def __next__(self):
        return self.cursor.__next__()

    @property
    def connection(self):
        return self.cursor.connection

    @property
    def description(self):
        return self.cursor.description

    # @property
    # def lastrowid(self):
    #     return self.cursor.lastrowid

    @property
    def rowcount(self):
        return self.cursor.rowcount



from .config_base import ConfigBase
from .exceptions import DatatypeException, DataException
from .managed_cursor import ManagedCursor
from .connection_base import ConnectionBase
from .mysql_connection import MySqlConnection
from .pgsql_connection import PgSqlConnection
from .sqlite_connection import SqliteConnection
from .session import Session, PersistentSession
from .session_factory import SessionFactory



from typing import Any

from .managed_cursor import ManagedCursor


class ConnectionBase(object):
    def __init__(self, connection_string: str):
        self.connection_string: str = connection_string
        self.database: str = ""
        self.connection: Any = None
        self.provider_name: str = ""
        self.cursor = None

    def __enter__(self):
        return self

    def __exit__(self, type, value, traceback):
        self.close()

    def start(self):
        ...

    def commit(self):
        ...

    def rollback(self):
        ...

    def execute(self, query: str, params: None) -> ManagedCursor:
        ...

    def execute_lastrowid(self, query: str, params: {}):
        ...

    def fetch(self, query: str, params=None) -> ManagedCursor:
        ...

    def new_cursor(self) -> Any:
        return self.connection.cursor()

    def close(self):
        ...
